// Main driver module for Arasan chess engine.
// Handles Winboard/xboard/UCI protocol.
// Copyright 1997-2012 by Jon Dart. All Rights Reserved.
//
#include "board.h"
#include "movegen.h"
#include "search.h"
#include "movearr.h"
#include "notation.h"
#include "scoring.h"
#include "chessio.h"
#include "bitboard.h"
#include "attacks.h"
#include "globals.h"
#include "log.h"
#include "util.h"
#include "calctime.h"
#include "eco.h"
#include "strpair.h"
#include "learn.h"
#include "boardio.h"
#include "legal.h"
#include "bitprobe.h"
extern "C"
{
#include <string.h>
#include <ctype.h>
}
#include "book.h"
#ifdef GAVIOTA_TBS
#include "gtb.h"
#endif
#ifdef NALIMOV_TBS
#include "tbprobe.h"
#endif
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <sys/timeb.h>

#include <string>
#include <iomanip>
using namespace std;

static int verbose = 0, post = 0;
static SearchController *searcher = NULL;

static Move last_move;
static string last_move_image;
static string game_pathname;
static ofstream *game_file = NULL;
static int32 time_left = 0;
static int32 opp_time = 0;
static float minutes = 5.0;
static int incr = 0; // Winboard increment
static int winc = 0; // UCI increment
static int binc = 0; // UCI increment
static int computer = 0;
static bool computer_plays_white = false;
static string opponent_name;
static bool ics = false;
static int forceMode = 0, analyzeMode = 0;
static int moves = 40;                            /* moves in a time control bloc */
static Board *ponder_board = NULL, *main_board = NULL;
static bool ponder_move_ok = false;
static Move predicted_move;
static int pondering = 0;
static Move ponder_move, best_move;
static Statistics stats, last_stats, ponder_stats;
static int time_target = 0;
static int num_pending = 0;
static int computer_rating = 0;
static int opponent_rating = 0;
#define MAX_PENDING 20
// stack of pending commands
static string pending[MAX_PENDING];
static bool doTrace = false;
static bool easy = false;
static int game_end = 0;
static int32 last_score = Constants::MATE;
static BookInfo book_inf;
static ECO *ecoCoder = NULL;
static string hostname;
static int xboard42 = 0;
static SearchType srctype = TimeLimit;
static int time_limit;
static int ply_limit;
static string start_fen;
static int tb_init = 0;
static int uci = 0;                               // non-zero for UCI mode
static int movestogo = 0;
static int ponderhit = 0, stopped = 0;
static int testing = 0;
// set true if waiting for "ponderhit" or "stop"
static int uciWaitState = 0;
static string test_file;
static int cpusSet = 0; // true if cmd line specifies -c
static int memorySet = 0; // true if cmd line specifies -H

#ifdef UCI_LOG
fstream ucilog;
#endif

// profile
uint64 total_nodes = 0L;
int total_correct = 0L;
int total_tests = 0L;
uint64 total_time = 0L;
int early_exit_plies = Constants::MaxPly;
int early_exit = 0;
int solution_move_count = 0;
time_t solution_time = 0;
uint64 solution_nodes = (uint64)0;
int last_iteration_depth = -1;
int iterations_correct = 0;
int moves_to_search = 1;
int flip = 0;
Move solution_moves[10];
bool avoid = false;
ArasanVector<int> solution_times;
static uint64 nodes_to_find_total;
static int depth_to_find_total;
static uint64 time_to_find_total;
static int max_depth;
static struct
{
   Move move;
   int value;
   time_t time;
   int depth;
   uint64 num_nodes;
} search_progress[256];

#ifdef _WIN32

static int polling_terminated = 0;
static HANDLE pollingThreadHandle;

static DWORD WINAPI inputPoll(void *x) {
   HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);
   if (doTrace) cout << "# starting poll thread" << endl;
   int offset = 0;
   int len = 0;
   char buf[1024];
   while (!polling_terminated) {
      DWORD dwRead;
      BOOL bSuccess = ReadFile(hStdin, buf, 1024, &dwRead, NULL);
      if (! bSuccess || dwRead == 0)
         break;
      Lock(input_lock);
      // copy to command buffer
      ASSERT(cmd_buf_count < CMD_BUF_SIZE);
      ASSERT(dwRead > 0 && dwRead < (DWORD)(CMD_BUF_SIZE-cmd_buf_count));
      if (dwRead) {
         if (doTrace) cout << "# got input: " << dwRead << " chars." << endl;
         if (cmd_buf_count+dwRead>CMD_BUF_SIZE) {
           if (doTrace) cout << "# warning: command buffer full!" << endl;
           dwRead-=(CMD_BUF_SIZE-cmd_buf_count);
         }
         strncpy(cmd_buf+cmd_buf_count,buf,dwRead);
         cmd_buf_count += dwRead;
      }
      Unlock(input_lock);
      // signal reader
      SetEvent(inputEvent);
   }
   return 0;
}
#endif

// forward declaration
static bool do_command(const char *buf, Board &board);

static void parseLevel(const char *cmd) {
  float time1, time2;
  if(sscanf (cmd,
             "%d %f:%f %d",
             &moves,
             &time1,
             &time2,
             &incr)==4) {
      minutes = time1 + time2/60;
  } else if(sscanf (cmd, "%d %f %d", 
                   &moves,
                   &minutes,
		   &incr)!=3) {
     cerr << "invalid level command, could not parse '" << cmd << "'" << endl;
     return;
  }
  // Winboard increment is in seconds, convert to our
  // internal value (seconds*100)
  incr *= 100;
}

static bool is_move(const Board &b,const char *buf) {
   if (isalpha(*buf) && isdigit(buf[1]) &&
      isalpha(buf[2]) && isdigit(buf[3]))
      return true;
   else if (strncmp(buf,"o-o",3)==0 || strncmp(buf,"O-O",3)==0 ||
      strncmp(buf,"o-o-o",5)==0 || strncmp(buf,"O-O-O",5) ==0)
      return true;
   else {
      return Notation::value(b,b.sideToMove(),buf) != NullMove;
   }
}


static Move text_to_move(const Board &board,const char *input) {
   Move rmove;

   // Winboard/xboard 4.2 and above support SAN. First try to
   // decode the move as a SAN string.
   //
   rmove = Notation::value(board,board.sideToMove(),input);
   if (!IsNull(rmove)) return rmove;              // success

   // Now try "old" coordinate-style notation, used in Winboard/
   // xboard before 4.2
   //
   string moveStr(input);

   if (moveStr == "o-o" ||
      moveStr == "o-o-o" ||
      moveStr == "O-O" ||
   moveStr == "O-O-O") {
      string::iterator it;
      for ( it=moveStr.begin() ; it < moveStr.end(); it++ ) {
         (*it) = toupper(*it);
      }
      rmove = Notation::moveValue(board,moveStr.c_str(),board.sideToMove());
   }
   else {
      Square start = SquareValue(input);
      Square dest = SquareValue(input+2);
      if (input[4]) {
         PieceType promotion =
            PieceCharValue(toupper(input[4]));
         rmove = CreateMove(board,start,dest,
            promotion);
      }
      else
         rmove = CreateMove(board,start,dest,Empty);
   }
   return rmove;
}

static int getIncrUCI(const ColorType side) {
    return side == White ? winc : binc;
}

static void delayedInitIfNeeded() {
   if (!tb_init) {
      delayedInit(); tb_init++;
#if defined(NALIMOV_TBS) || defined(GAVIOTA_TBS)
      string path;
#ifdef GAVIOTA_TBS
      if (options.search.tablebase_type == Options::GAVIOTA_TYPE) {
          path = options.search.gtb_path;
      }
#endif
#ifdef NALIMOV_TBS
      if (options.search.tablebase_type == Options::NALIMOV_TYPE) {
          path = options.search.nalimov_path;
      }
#endif
      if (EGTBMenCount)
          cerr << "found " << EGTBMenCount << "-man " <<
              options.search.tablebase_type << " tablebases in directory " <<
              path  << endl;
#endif
   }
}

static bool accept_draw(Board &board) {
   if (doTrace)
      cout << "# in accept_draw" << endl;
   // Code to handle draw offers.  This is somewhat experimental.
   int rating_diff = opponent_rating - computer_rating;
   // ignore draw if we have just started searching
   if (last_score == Constants::MATE)
      return false;
   // If it's a 0-increment game and the opponent has < 1 minute,
   // and we have more time, decline
   int inc = uci ? getIncrUCI(board.oppositeSide()) : incr;
   if (!computer && inc == 0 && opp_time < 6000 && time_left > opp_time)
      return false;
   // See if we do not have enough material to mate
   const ColorType side = computer_plays_white ? White : Black;
   const Material &ourmat = board.getMaterial(side);
   const Material &oppmat = board.getMaterial(OppositeColor(side));
   if (ourmat.noPawns() && ourmat.value() <= KING_VALUE + BISHOP_VALUE) {
      // We don't have mating material
      if (doTrace)
         cout << "# no mating material, accept draw" << endl;
      return true;
   }
   // accept a draw in pawnless endings with even material, unless
   // our score is way positive
   if (ourmat.noPawns() && oppmat.noPawns() &&
      (Util::Abs(ourmat.value() - oppmat.value()) < (int)PAWN_VALUE/2) &&
      last_score < (int)PAWN_VALUE) {
      if (doTrace)
         cout << "# pawnless ending, accept draw" << endl;
      return true;
   }
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
   const Material &wMat = board.getMaterial(White);
   const Material &bMat = board.getMaterial(Black);
   if(options.search.use_tablebases &&
      wMat.men() + bMat.men() <= EGTBMenCount) {
      if (doTrace)
         cout << "# checking tablebases .." << endl;
      // accept a draw when the tablebases say it's a draw
      int tbscore;
#ifdef GAVIOTA_TBS
      if (options.search.tablebase_type == Options::GAVIOTA_TYPE) {
          if (GaviotaTb::probe_tb(board,tbscore,0,1) && tbscore == 0) {
              if (doTrace)
                  cout << "# tablebase score says draw" << endl;
              return true;
          }
      }
#endif
#ifdef NALIMOV_TBS
      if (options.search.tablebase_type == Options::NALIMOV_TYPE) {
          if (NalimovTb::probe_tb(board,tbscore,0) && tbscore == 0) {
              if (doTrace)
                  cout << "# tablebase score says draw" << endl;
              return true;
          }
      }
#endif
   }
#endif
   if (opponent_rating == 0)
      return false;
   // accept a draw if our score is negative .. how much negative
   // depends on opponent rating.
   if (doTrace)
      cout << "# checking draw score .." << endl;
   ColorType tmp = board.sideToMove();
   board.setSideToMove(side);
   int draw_score = searcher->drawScore(board);
   board.setSideToMove(tmp);
   int threshold = (3*PAWN_VALUE)/4 -(PAWN_VALUE*rating_diff)/200;
   if (rating_diff > 0)
      threshold = Util::Max(-PAWN_VALUE,threshold);
   else
      threshold = Util::Min(3*PAWN_VALUE,threshold);
   if (doTrace) {
      cout << "# rating_diff = " << rating_diff << endl;
      cout << "# draw_score = " << draw_score << endl;
      cout << "# last_score = " << last_score << endl;
      cout << "# threshold = " << threshold << endl;
   }
   return draw_score-last_score > threshold;
}


static void do_help() {
#ifndef _WIN32
   if (!isatty(1)) return;
#endif
   cout << "analyze:         enter Winboard analyze mode" << endl;
   cout << "black:           set computer to play Black" << endl;
   cout << "bk:              show book moves" << endl;
   cout << "computer:        used to indicate the opponent is a computer" << endl;
   cout << "draw:            offer a draw" << endl;
   cout << "easy:            disable pondering" << endl;
   cout << "edit:            enter Winboard edit mode" << endl;
   cout << "force:           disable computer moving" << endl;
   cout << "go:              start searching" << endl;
   cout << "hard:            enable pondering" << endl;
   cout << "hint:            compute a hint for the current position" << endl;
   cout << "ics <hostname>:  set the name of the ICS host" << endl;
   cout << "level <a b c>:   set the time control:" << endl;
   cout << "  a -> moves to time control" << endl;
   cout << "  b -> minutes per game" << endl;
   cout << "  c -> increment in seconds" << endl;
   cout << "name <string>:   set the name of the opponent" << endl;
   cout << "new:             start a new game" << endl;
   cout << "nopost:          disable output during search" << endl;
   cout << "otim <int>:      set opponent time remaining (in centiseconds)" << endl;
   cout << "post:            show output during search" << endl;
   cout << "quit:            terminate the program" << endl;
   cout << "remove:          back up a full move" << endl;
   cout << "resign:          resign the current game" << endl;
   cout << "result <string>: set the game result (0-1, 1/2-1/2 or 1-0)" << endl;
   cout << "sd <x>:          limit thinking to depth x" << endl;
   cout << "setboard <FEN>:  set board to a specified FEN string" << endl;
   cout << "st <x>:          limit thinking to x centiseconds" << endl;
   cout << "time <int>:      set computer time remaining (in centiseconds)" << endl;
   cout << "undo:            back up a half move" << endl;
   cout << "white:           set computer to play White" << endl;
   cout << "test <file> <seconds> <early exit> <# moves> <-v> <outfile>: "<< endl;
   cout << "   - run an EPD testsuite" << endl;
   cout << "eval <file>:     evaluate a FEN position." << endl;
}


static void save_game() {
   if (uci) return;                               // not supported
   if (doTrace) cout << "# in save_game" << endl;
   if (gameMoves->num_moves() == 0 || !options.store_games) {
      if (doTrace) cout << "# out of save_game" << endl;
      return;
   }
   if (options.learning.result_learning) {
      const Log::GameResult result = theLog->getResult();
      if (result != Log::Incomplete)
         learnResult(computer_plays_white ? White: Black,
         result,opponent_rating-computer_rating);
   }
   if (game_file) {
      ArasanVector<StringPair> headers;
      string opening_name, eco;
      if (ecoCoder) {
         if (doTrace) cout << "# calling classify" << endl;
         ecoCoder->classify(*gameMoves,eco,opening_name);
         headers.append(StringPair("ECO",eco.c_str()));
      }
      static char timec[50];
      static char crating[15];
      static char orating[15];
      sprintf(crating,"%d",computer_rating);
      sprintf(orating,"%d",opponent_rating);

      if (hostname.length() > 0) {
         headers.append(StringPair("Site",hostname.c_str()));
      }

      if (computer_plays_white) {
         headers.append(StringPair("Black",
            opponent_name.length() > 0 ? opponent_name.c_str() : "?"));
         if (computer_rating)
            headers.append(StringPair("WhiteElo",crating));
         if (opponent_rating)
            headers.append(StringPair("BlackElo",orating));
      }
      else {
         headers.append(StringPair("White",
            opponent_name.length() > 0 ? opponent_name.c_str() : "?"));
         if (opponent_rating)
            headers.append(StringPair("WhiteElo",orating));
         if (computer_rating)
            headers.append(StringPair("BlackElo",crating));
      }
      if (start_fen.size()) {
         // we had a non-standard starting position for the game
          headers.append(StringPair("FEN",start_fen.c_str()));
      }
      sprintf(timec,"%d+%d",(int)(minutes*60),incr/100);
      headers.append(StringPair("TimeControl",timec));
      string result;
      theLog->getResultAsString(result);
      ChessIO::store_pgn(*game_file, *gameMoves,
         computer_plays_white ? White : Black,
         result.c_str(),
         headers);
      // free headers
      headers.removeAll();
   }
   if (doTrace) cout << "# out of save_game" << endl;
}


static int calc_extra_time(const ColorType side) {
   // Determine if we are allowed to use any time beyond
   // the normal time target.
   int inc = (uci ? getIncrUCI(side) : incr);
   if (srctype == FixedTime || srctype == FixedDepth)
      return 0;
   if (time_left < 1500)
      return 0;                                   // no extra time
   else if (time_target <= 100)
      return 100;                                 // lightning game
   else if ((inc == 0 && time_left > time_target*10) ||
      (inc > 0 && time_left > time_target*6))
      return time_target;
   else
      return 0;
}


// handle ^C interrupts (sent by winboard)
static void CDECL sigHandler(int) {
   cout << "got ^C" << endl;
   if (searcher) {
      searcher->terminateNow();
   }
   signal(SIGINT,sigHandler);
}


// Produce a text string representing the move, in a form
// the GUI understands
static void move_image(const Board &board, Move m, ostream &buf, int uci) {
   if (uci) {
      Notation::UCIMoveImage(m,buf);
   }
   else {
      if (TypeOfMove(m) == KCastle) {
          buf << "O-O";
      }
      else if (TypeOfMove(m) == QCastle) {
          buf << "O-O-O";
      }
      else {
         buf << FileImage(StartSquare(m));
         buf << RankImage(StartSquare(m));
         buf << FileImage(DestSquare(m));
         buf << RankImage(DestSquare(m));
         if (TypeOfMove(m) == Promotion) {
            // N.b. ICS requires lower case.
            buf << (char)tolower((int)PieceImage(PromoteTo(m)));
         }
      }
   }
}


static void print_nodes(uint64 nodes, ostream &out) {
   if (nodes >= 1000000) {
      cout << (float)(nodes)/1000000.0 << "M";
   }
   else if (nodes >= 1000) {
      cout << (float)(nodes)/1000.0 << "K";
   }
   else
      cout << nodes;
}


static struct MultiPVEntry
{
   int depth;
   int score;
   time_t time;
   uint64 nodes;
   uint64 tb_hits;
   string best_line_image;
   Move best;
   MultiPVEntry() {}
   MultiPVEntry(const Statistics &stats)
      : depth(stats.depth),score(stats.value),
      time(stats.elapsed_time),nodes(stats.num_nodes),
      tb_hits(stats.tb_hits) {
      best_line_image = stats.best_line_image;
      best = stats.best_line[0];
   }
} multi_pvs[MAX_PV];

static void uciOut(int depth, int score, time_t time,
uint64 nodes, uint64 tb_hits, const string &best_line_image, int multipv) {
   stringstream s;
   s << "info";
   if (multipv) {
      s << " multipv " << multipv;
   }
   s << " depth " << depth << " score ";
   Scoring::printScoreUCI(score,s);
   s << " time " << time*10L << " nodes " << nodes;
   if (time>30) s << " nps " << (long)((100L*nodes)/time);
   if (tb_hits) {
      s << " tbhits " << tb_hits;
   }
   s << " hashfull " << Hash::pctFull();
   if (best_line_image.length()) {
      s << " pv ";
      s << best_line_image;
   }
   cout << s.str() << endl;
   if (doTrace) {
      theLog->write(s.str().c_str()); theLog->write_eol();
   }
#ifdef UCI_LOG
   ucilog << s.str() << endl;
#endif   
}


static void uciOut(const Statistics &stats) {
   uciOut(stats.depth,stats.value,stats.elapsed_time,
      stats.num_nodes,stats.tb_hits,
      stats.best_line_image,0);
}


static void post_output(const Statistics &stats, int force) {
   last_score = stats.value;
   int score = stats.display_value;
   if (verbose) {
      if (uci) {
         if (stats.elapsed_time) {
            if (options.search.multipv > 1) {
               // accumulate multiple pvs until we are ready to output them
               ASSERT(stats.multipv_count<MAX_PV);
               multi_pvs[stats.multipv_count] = MultiPVEntry(stats);
#ifdef UCI_LOG
               ucilog << "multipv " << stats.multipv_count+ 1 << endl;
#endif
               if (stats.multipv_count+1 == options.search.multipv) {
                  for (int i = 0; i < options.search.multipv; i++) {
                     uciOut(multi_pvs[i].depth,multi_pvs[i].score,
                        multi_pvs[i].time,multi_pvs[i].nodes,
                        multi_pvs[i].tb_hits,
                        multi_pvs[i].best_line_image,
                        i+1);
                  }
               }
            }
            else {
               uciOut(stats);
            }
         }
      }
   }
   else if (post) {
      // "post" output for Winboard
#ifdef _WIN32
      printf("%2u%c %6d %6ld %8I64u %s\n",
#else
      printf("%2u%c %6d %6ld %8llu %s\n",
#endif
         stats.depth,' ',
         score,
         (long)stats.elapsed_time,
         stats.num_nodes,
         stats.best_line_image.c_str());
         fflush(stdout);
   }
}


static void CDECL post_output(const Statistics &stats) {
   post_output(stats,0);
}


static int solution_match(Move result) {
   for (int i = 0; i < solution_move_count; ++i) {
      if (MovesEqual(solution_moves[i],result)) {
         return 1;
      }
   }
   return 0;
}


// this function is called with "post" results during a test
// suite run.
static void CDECL post_test(const Statistics &stats)
{
   Move best = stats.best_line[0];
   if (!IsNull(best) && max_depth < 256) {
      int ply = stats.depth-1;
         search_progress[max_depth].move = best;
         search_progress[max_depth].value = stats.value;
         search_progress[max_depth].time = stats.elapsed_time;
         search_progress[max_depth].depth = ply;
         search_progress[max_depth].num_nodes = stats.num_nodes;
         max_depth++;
   }
   if (avoid) {
      // note: doesn't handle multiple "am" moves
      int ok = !MovesEqual(solution_moves[0],best);
      if (ok) {
         if (solution_time == -1) {
            solution_time = stats.elapsed_time;
               solution_nodes = stats.num_nodes;
         }
      }
      else
         solution_time = -1;
         return;
   }
   else {
      if (solution_match(best)) {
         if ((int)stats.depth > last_iteration_depth) {
            // Wait 2 sec before counting iterations correct, unless
            // we found a mate
            if (stats.elapsed_time >= 200 ||
               stats.value > Constants::MATE-200)
               ++iterations_correct;
               last_iteration_depth = stats.depth;
         }
         if (iterations_correct >= early_exit_plies)
            early_exit = 1;
         if (solution_time == -1) {
            solution_time = stats.elapsed_time;
               solution_nodes = stats.num_nodes;
         }
         return;
      }
      else
         solution_time = -1;                      // not solved yet, or has moved off solution
   }
   iterations_correct = 0;
}


static int CDECL terminate(const Statistics &stats)
{
   post_test(stats);
   return early_exit;
}


// handle editing the chessboard (Winboard mode)
static void edit_board(istream &fin, Board &board)
{
    for (int i = 0; i < 64; i++)
        board.setContents(EmptyPiece,i);
    static char buf[80];
    static const char pieces[]="PNBRQKpnbrqk";
    ColorType side = White;
    while (fin.getline(buf,80)) {
        if (strcmp(buf,"white") == 0)
            side = White;
        else if (strcmp(buf,"black")==0)
            side = Black;
        else if (*buf == '#') {
            for (int i = 0; i < 64; i++)
                board.setContents(EmptyPiece,i);
        }
        else if (strcmp(buf,"c")==0)
            side = OppositeColor(side);
        else if (strcmp(buf,".")==0)
            break;
        else {
            if (strchr(pieces,buf[0])) {
                char c = tolower(buf[0]);
               Piece p = EmptyPiece;
               switch (c) {
               case 'p':
                   p = MakePiece(Pawn,side);
                   break;
               case 'n':
                   p = MakePiece(Knight,side);
                   break;
               case 'b':
                   p = MakePiece(Bishop,side);
                   break;
               case 'r':
                   p = MakePiece(Rook,side);
                   break;
               case 'q':
                   p = MakePiece(Queen,side);
                   break;
               case 'k':
                   p = MakePiece(King,side);
                   break;
               default:
                   break;
                }

                if (p != EmptyPiece) {
                    Square sq = SquareValue(buf+1);
                    if (!IsInvalid(sq))
                        board.setContents(p,sq);
                }
            }
        }
    }
    board.setSecondaryVars();
    // edit doesn't set the castle status, so try to deduce it
    // from the piece positions
    if (board.kingSquare(White) == E1) {
        if (board[A1] == WhiteRook &&
            board[H1] == WhiteRook)
            board.setCastleStatus(CanCastleEitherSide,White);
        else if (board[A1] == WhiteRook)
            board.setCastleStatus(CanCastleQSide,White);
        else if (board[H1] == WhiteRook)
            board.setCastleStatus(CanCastleKSide,White);
    }
    else
        board.setCastleStatus(CantCastleEitherSide,White);
    if (board.kingSquare(Black) == E8) {
        if (board[A8] == BlackRook &&
            board[H8] == BlackRook)
            board.setCastleStatus(CanCastleEitherSide,Black);
        else if (board[A8] == BlackRook)
            board.setCastleStatus(CanCastleQSide,Black);
        else if (board[H8] == BlackRook)
            board.setCastleStatus(CanCastleKSide,Black);
    }
    else
        board.setCastleStatus(CantCastleEitherSide,Black);
}


static void ponder(Board &board, Move move, Move predicted_reply,int uci)
{
    ponder_move_ok = false;
    ponder_move = NullMove;
    ponder_stats.clear();
    if (doTrace) {
        cout << "# predicted reply = ";
        MoveImage(predicted_reply,cout);
        cout << endl;
    }
    if (uci || (!IsNull(move) && !IsNull(predicted_reply))) {
        if (!uci) {
            predicted_move = predicted_reply;
            // We have already set up the ponder board with the board
            // position after our last move. Now make the move we predicted.
            //
            BoardState previous_state = ponder_board->state;
            ASSERT(legalMove(*ponder_board,predicted_reply));
            ponder_board->doMove(predicted_reply);
            //
            // We must also add this move to the global move list,
            // otherwise repetition detection will be broken.
            //
            gameMoves->add_move(board,previous_state,predicted_reply,"");
            // Start a search for the reply
        }
        book_inf.clear();
        if (options.book.book_enabled && openingBook)
            ponder_move = openingBook->book_move(board,book_inf);
        else
            ponder_move = NullMove;
        if (IsNull(ponder_move)) {
            pondering++;
            if (doTrace)
                cout << "# starting to ponder" << endl;
            if (srctype == FixedDepth)
                ponder_move = searcher->findBestMove(
                uci ? *main_board : *ponder_board,
                srctype,
                0,
                0,
                ply_limit, true, uci,
                ponder_stats,
                (doTrace) ? Trace : Silent);
            else
                ponder_move = searcher->findBestMove(
                uci ? *main_board : *ponder_board,
                FixedTime,
                99999,        /* time limit (999 seconds) */
                0,            /* extra time allowed */
                Constants::MaxPly,           /* ply limit */
                true,         /* background */
                uci,
                ponder_stats,
                (doTrace) ? Trace : Silent);
            pondering--;
            if (doTrace) {
                cout << "# done pondering" << endl;
                cout << "# ponder move = ";
                MoveImage(ponder_move,cout);
                cout << endl;
            }
        }
        else                                        // we got a book move
            ponder_move = NullMove;
        // Clean up the global move array, if we got no ponder hit.
        if (!uci && gameMoves->length() > 0 && strlen(gameMoves->last().image())==0)
            gameMoves->remove_move();
    }
    if (doTrace) cout << "# out of ponder()" << endl;
}


// Search using the current board position.
//
static Move search(SearchController *searcher, Board &board, Statistics &stats, bool infinite)
{
    last_stats.clear();
    last_score = Constants::MATE;
    book_inf.clear();
    ponder_move = NullMove;
    if (doTrace) cout << "# in search()" << endl;

    Move move = NullMove;
    if (!infinite && options.book.book_enabled && !testing && openingBook) {
        move = openingBook->book_move(board,book_inf);
    }

    if (IsNull(move)) {
        // no book move
        stats.clear();
        TalkLevel level = Silent;
        if (verbose && !uci) level = Debug;
        else if (doTrace) level = Trace;
        else level = Silent;
        if (srctype == FixedDepth) {
            move = searcher->findBestMove(board,
                srctype,
                0,
                0,
                ply_limit, false, uci,
                stats,
                level);
        }
        else {
            if (infinite) {
                time_target = INFINITE_TIME;
            } else {
                time_target =
                    (srctype == FixedTime) ? time_limit :
                    (uci ? calcTimeLimit(movestogo, 
                                         getIncrUCI(board.sideToMove()),
                                         time_left, opp_time, false, doTrace)
                     : calcTimeLimit(moves, minutes, incr, time_left, opp_time, false, doTrace));
            }
            if (doTrace) {
                cout << "# entering search, time_target = " << time_target << endl;
                cout << "# xtra time = " << calc_extra_time(board.sideToMove()) << endl;
            }
            move = searcher->findBestMove(board,
                srctype,
                time_target,
                                          calc_extra_time(board.sideToMove()),
                Constants::MaxPly, false, uci,
                stats,
                level);
        }
#ifdef UCI_LOG
        ucilog << "after findBestMove" << (flush) << endl;
#endif
        if (doTrace) {
            cout << "# search done : move = ";
            MoveImage(move,cout);
            cout << endl;
        }
        last_stats = stats;
        if (!forceMode) {
            if (testing)
                post_test(stats);
            else
                post_output(stats,1);
        }
    }
    else {
        if (ics || uci) {
            Move choices[20];
            int scores[20];
            int moveCount = 0;
            if (options.book.book_enabled && openingBook) {
                moveCount = openingBook->book_moves(board,choices,scores,20);
            }
            stringstream s;
            if (uci)
                s << "info string book moves: (";
            else if (computer)
                s << "tellics kibitz book moves: (";
            else
                s << "tellics whisper book moves: (";
            for (int i=0;i<moveCount;i++) {
                Notation::image(board,choices[i],s);
                s << ' ';
                s << scores[i];
                if (i < moveCount-1)
                    s << ", ";
            }
            s << "), choosing ";
            Notation::image(board,move,s);
            cout << s.str() << endl;
        }
        stats.clear();
    }
    return move;
}


static void send_move(Board &board, Move &move, Statistics
                      &stats) {
   last_move = move;
   last_stats = stats;
   ColorType sideToMove = board.sideToMove();
#ifdef UCI_LOG
   ucilog << "in sendmove move=";
   MoveImage(move,ucilog);
   ucilog << endl << (flush);
#endif  
   if (!IsNull(move)) {
       int rep_count = board.repCount();
       if (!uci && rep_count >= 3) {
           // It is possible the opponent has not claimed the
           // draw by repetition. If it's a draw before we even
           // move, don't make a move, just claim the draw
           cout << "1/2-1/2 {Repetition}" << endl;
           // Wait for Winboard to send a "result" command before
           // actually concluding it's a draw.
           return;
       }
       stringstream img;
       Notation::image(board,last_move,img);
       last_move_image = img.str();
       theLog->add_move(board,last_move,last_move_image,&last_stats,&book_inf,true);
       // Perform book and/or position learning (if enabled):
       learn(board,rep_count);
       stringstream movebuf;
       move_image(board,last_move,movebuf,uci);

       BoardState previous_state = board.state;
       board.doMove(last_move);
       gameMoves->add_move(board,previous_state,last_move,last_move_image);

       delete ponder_board;
       ponder_board = new Board(board);

       if (uci) {
#ifdef UCI_LOG
           ucilog << "engine: bestmove " << movebuf << (flush) << endl;
#endif  
           cout << "bestmove " << movebuf.str();
           if (!IsNull(stats.best_line[1])) {
               stringstream ponderbuf;
               move_image(board,stats.best_line[1],ponderbuf,uci);
               cout << " ponder " << ponderbuf.str();
           }
           cout << endl;
       }
       else if (xboard42) {
           cout << "move " << movebuf.str() << endl;
       }
       else {
           cout << gameMoves->num_moves()/2 << ". ... ";
           cout << movebuf.str() << endl;
       }
       cout << (flush);
   }
   else {
       cout << "# warning : move is null" << endl;
   }
   if (ics && time_target >= 300 && stats.num_moves > 0) {
       if (computer)
           cout << "tellics kibitz ";
       else
           cout << "tellics whisper ";
       cout << "time=" << fixed << setprecision(2) <<
          (float)last_stats.elapsed_time/100.0 << " sec. score=";
       Scoring::printScore(last_stats.display_value,cout);
       cout << " depth=" << last_stats.depth;
       if (last_stats.elapsed_time > 0) {
           cout << " nps=";
           last_stats.printNPS(cout);
       }
       if (last_stats.tb_hits) {
           cout << " egtb=" << last_stats.tb_hits << '/' << last_stats.tb_probes;
       }
#if defined(SMP_STATS)
       if (last_stats.samples && options.search.ncpus>1) {
           cout << " cpu=" << fixed << setprecision(2) <<
                (100.0*last_stats.threads)/((float)last_stats.samples) << '%';
       }
#endif
       if (last_stats.best_line_image.length() && !game_end) {
           cout << " pv: " << last_stats.best_line_image;
       }
       cout << endl;
   }
   if (uci) return; // With UCI, GUI is in charge of game end detection
   // check for checkmate first, before draw (avoids problem
   // when last move was mate but also hits 50 move counter).
   if (last_stats.value >= Constants::MATE-1) {
       if (doTrace) cout << "# last_score = mate" << endl;
       if (sideToMove == White) {
           theLog->setResult("1-0");
           cout << "1-0 {White mates}" << endl;
       }
       else {
           theLog->setResult("0-1");
           cout << "0-1 {Black mates}" << endl;
       }
       game_end = 1;
   }
   else if (last_stats.state == Checkmate) {
       if (doTrace) cout << "# state = Checkmate" << endl;
       if (sideToMove == White) {
           theLog->setResult("0-1");
           cout << "0-1 {Black mates}" << endl;
       }
       else {
           theLog->setResult("1-0");
           cout << "1-0 {White mates}" << endl;
       }
       game_end = 1;
   }
   else if (last_stats.state == Stalemate) {
       cout << "1/2-1/2 {Stalemate}" << endl;
       theLog->setResult("1/2-1/2 {stalemate}");
       game_end = 1;
   }
   else if (board.state.moveCount >= 100) {
       cout << "1/2-1/2 {50 move rule}" << endl;
       theLog->setResult("1/2-1/2 {50 move draw}");
       if (doTrace) cout << "# 50 move draw" << endl;
       game_end = 1;
   }
   else if (Scoring::materialDraw(board)) {
       cout << "1/2-1/2 {Insufficient material}" << endl;
       theLog->setResult("1/2-1/2 {insufficient material}");
       if (doTrace) cout << "# material draw" << endl;
       game_end = 1;
   }
   else if (Scoring::repetitionDraw(board)) {
       cout << "1/2-1/2 {Repetition}" << endl;
       theLog->setResult("1/2-1/2 {repetition}");
       game_end = 1;
   }
   else if (stats.state == Draw) {
       // We can get here if our opponent does not claim the
       // repetition draw.
       if (doTrace) cout << "# opponent didn't claim repetition?" << endl;
       cout << "1/2-1/2 {Repetition}" << endl;
   }
   else if (last_stats.state == Resigns) {
       // Don't resign a zero-increment game if the opponent is short
       // on time
       if (!(incr == 0 && opp_time < 2000)) {
           // Winboard passes the resign command to ICS, but ignores it
           // itself.
           if (computer_plays_white) {
               theLog->setResult("0-1 {White resigns}");
               cout << "0-1 {White resigns}" << endl;
           }
           else {
               theLog->setResult("1-0 {Black resigns}");
               cout << "1-0 {Black resigns}" << endl;
           }
           game_end = 1;
       }
       else {   // reset flag - we're not resigning
           stats.end_of_game = 0;
           game_end = 0;
       }
   }
}


// Find best move using the current board position.
//
static Move analyze(SearchController &searcher, Board &board, Statistics &stats)
{
    last_stats.clear();
    last_score = Constants::MATE;

    stats.clear();
    if (doTrace)
        cout << "# entering analysis search" << endl;
    Move move = searcher.findBestMove(board,
        FixedTime,
        INFINITE_TIME,
        0,
        Constants::MaxPly, false, uci,
        stats,
        Whisper);
    if (doTrace) {
        cout << "# search done : move = ";
        MoveImage(move,cout);
        cout << endl;
    }

    last_stats = stats;
    post_output(stats);

    return move;
}


static void doHint() {
    // try book move first
    Move moves[20];
    int scores[20];
    unsigned count = 0;
    if (options.book.book_enabled && openingBook)
        count = openingBook->book_moves(*main_board,moves,scores,20);
    if (count > 0) {
        if (count == 1)
            cout << "Book move: " ;
        else
            cout << "Book moves: ";
        for (unsigned i = 0; i<count; i++) {
            Notation::image(*main_board,moves[i],cout);
            if (i<count-1) cout << ' ';
        }
        cout << endl;
        return;
    }
    else {
        // no book move, see if we have a ponder move
        const string &img = last_stats.best_line_image;
        if (img.length()) {
            string::const_iterator it = img.begin();
            while (it != img.end() && !isspace(*it)) it++;
            string last_move;
            if (it != img.end()) {
                it++;
                while (it != img.end() && !isspace(*it)) last_move += *it++;
            }
            if (last_move.length()) {
                cout << "Hint: " << last_move << endl;
                return;
            }
        }
    }
    // no ponder move or book move. If we are already pondering but
    // have no ponder move we could wait a while for a ponder result,
    // but we just return for now.
    if (pondering) return;
    if (doTrace) cout << "# computing hint" << endl;

    Statistics tmp;
    // do low-depth search for hint move
    Move move = searcher->findBestMove(*main_board,
        FixedDepth,
        0,
        0,
        4, false, uci,
        tmp,
        (doTrace) ? Trace : Silent);
    if (!IsNull(move)) {
        cout << "Hint: ";
        Notation::image(*main_board,move,cout);
        cout << endl;
    }
}


void add_pending(const char *buf)
{
    ASSERT(num_pending < MAX_PENDING);
    if (doTrace) cout << "# adding to pending list " << buf << ", list size=" << num_pending << endl;
    pending[num_pending] = buf;
    ++num_pending;
}


static bool do_pending(Board &board)
{
    int limit = num_pending;
    if (doTrace) cout << "# in do_pending, list size=" << limit << endl;
    for (int i = 0; i < limit; i++) {
        --num_pending;
        bool terminate;
        if (doTrace)
            cout << "# pending command: " << pending[i] << endl;
#ifdef UCI_LOG
        ucilog << "gui (pending): " << pending[i] << (flush) << endl;
#endif  
        terminate = !do_command(pending[i].c_str(),board);
        if (terminate) {
            int k = 0;
            for (int j = i+1; j<limit;j++) {
                pending[k++] = pending[j];
            }
            if (doTrace) {
                cout << "# still pending:" << endl;
                for (int k=0;k<num_pending;k++)
                    cout << "# " << pending[k].c_str() << endl;
            }
            if (doTrace) cout << "# out of do_pending, list size=" << num_pending << endl;
            return false;
        }
    }
    if (doTrace) cout << "# out of do_pending, list size=" << num_pending << endl;
    return true;
}


//
// Do all pending commands in stack. Return 1 if "new" command
// was executed, 2 if "quit" was seen.
//
static int do_all_pending(Board &board)
{
    int limit = num_pending;
    int retVal = 0;
    if (doTrace) cout << "# in do_all_pending" << endl;
    for (int i = 0; i < limit; i++) {
        --num_pending;
        if (doTrace)
            cout << "# pending command(a): " << pending[i] << endl;
#ifdef UCI_LOG
        ucilog << "do_all_pending: " << pending[i] << (flush) << endl;
#endif
        if (strncmp(pending[i].c_str(),"new",3) == 0)
            retVal = 1;
        else if (strncmp(pending[i].c_str(),"quit",4) == 0) {
            retVal = 2;
            if (doTrace) cout << "# out of do_all_pending, list size=" << num_pending << endl;
            break;
        }
        do_command(pending[i].c_str(),board);
    }
    if (doTrace) cout << "# out of do_all_pending, list size=" << num_pending << endl;
    return retVal;
}


//
// Check for user move, resign or result in pending stack
//
static int check_pending(Board &board)
{
    int limit = num_pending;
    int i;
    for (i = 0; i < limit; i++) {
        const char *cmd = pending[i].c_str();
        if ((strncmp(cmd,"result",6) == 0) ||
            (strncmp(cmd,"new",3) == 0) ||
            (strncmp(cmd,"quit",4) == 0) ||
            (strncmp(cmd,"resign",6) == 0)) {
                if (doTrace) cout << "# game end signal in pending stack" << endl;
                for (int j=0;j<i;j++) {
                    pending[j]=pending[j+i];
                    num_pending--;
                    assert(num_pending>=0);
                }
                return 1;                                // game end
        }
        else if (strncmp(cmd,"usermove",8) == 0 || is_move(board,cmd)) {
            if (doTrace) cout << "# move in pending stack" << endl;
            for (int j=0;j<i;j++) {
                pending[j]=pending[j+i];
                num_pending--;
                assert(num_pending>=0);
            }
            return 2;
        }
        else {                                      // might as well execute this
#ifdef UCI_LOG
            ucilog << "calling do_command from check_pending" << (flush) << endl;
#endif
            do_command(cmd,board);
        }
    }
    for (int j=0;j<i;j++) {
        pending[j]=pending[j+i];
        num_pending--;
        assert(num_pending>=0);
    }
    return 0;
}

static void analyze(Board &board)
{
    while (analyzeMode) {
        Move reply = analyze(*searcher,board,stats);
        if (check_pending(board)) {
            // user move, do the command then come back to this loop
            if (doTrace) cout << "# exiting analysis loop" << endl;
            break;
        }
        if (do_all_pending(board)==2) return;
        if (reply && !IsNull(reply) && !game_end && analyzeMode) {
            // we got a reply
            if (doTrace) {
                cout << "# analyze result: ";
                MoveImage(reply,cout);
                cout << endl;
            }
            post_output(stats);
            if (Util::Abs(stats.value) >= Constants::MATE-200) {
                // We have a mate score. There is no point analyzing further.
                break;
            }
        }
        else {                                      // no analysis move
            if (doTrace) cout << "# no move returned from analyze" << endl;
            break;
        }
    }
}


static void undo( Board &board)
{
    if (theLog->current() < 1) return;             // ignore "undo"

    board.undoMove((*theLog)[theLog->current()-1].move(),
        (*theLog)[theLog->current()-1].state());
    theLog->back_up();
    gameMoves->remove_move();
    last_stats.clear();
    if (theLog->current()) {
        last_move = (*theLog)[theLog->current()-1].move();
        last_move_image = (*theLog)[theLog->current()-1].image();
    }
    else {
        last_move = NullMove;
        last_move_image.clear();
    }
    // In case we have backed up from the end of the game, reset
    // the "game end" flag.
    game_end = 0;
}


// Execute a move made by the opponent or in "force" mode.
static void execute_move(Board &board,Move m)
{
    if (doTrace) {
        cout << "# execute_move: ";
        MoveImage(m,cout);
        cout << endl;
    }
    last_move = m;
    stringstream img;
    Notation::image(board,m,img);
    last_move_image = img.str();
    theLog->add_move(board,m,last_move_image,NULL,NULL,true);
    BoardState previous_state = board.state;
    board.doMove(m);
    // If our last move added was the pondering move, replace it
    if (gameMoves->length() > 0 && strlen(gameMoves->last().image())==0)
        gameMoves->remove_move();
    gameMoves->add_move(board,previous_state,m,last_move_image);
}

static int setIntOption(const string &value, int &dest) {
    stringstream buf(value);
    int tmp;
    buf >> tmp;
    if (!buf.bad() && !buf.fail()) {
        dest = tmp;
        return 1;
    }
    return 0;
}

static void setCheckOption(const string &value, int &dest) {
    stringstream buf(value);
    int tmp;
    buf >> tmp;
    if (!buf.bad() && !buf.fail()) {
        dest = tmp != 0;
    }
}

static void processWinboardOptions(const char *buf) {
    string name, value;
    string args(buf);
    size_t eq = args.find("=");
    if (eq == string::npos) {
        // no value
        name = args;
    } else {
        name = args.substr(0,eq);
        value = args.substr(eq+1);
    }
    // trim spaces
    name = name.erase(0 , name.find_first_not_of(' ') );
    name = name.erase( name.find_last_not_of(' ') + 1);
    value = value.erase(0, value.find_first_not_of(' '));
    value = value.erase(value.find_last_not_of(' ') + 1);
    // handle option settings
    if (name == "Book selectivity") {
        setIntOption(value,options.book.selectivity);
    } else if (name == "Can resign") {
        setCheckOption(value,options.search.can_resign);
    } else if (name == "Score learning") {
        setCheckOption(value,options.learning.score_learning);
    } else if (name == "Result learning") {
        setCheckOption(value,options.learning.result_learning);
    } else if (name == "Position learning") {
        setCheckOption(value,options.learning.position_learning);
    }
}


// Handle a command received while searching.  terminate is set
// true if the search should stop.  Some commands are executed,
// some are placed on the pending stack to be executed after
// the search completes.
//
static void check_command(char *buf, int &terminate)
{
    if (doTrace)
        cout << "# check_command: " << buf << endl;
    if (doTrace && uci) {
        theLog->write("check_command: "); theLog->write(buf); theLog->write_eol();
    }
#ifdef UCI_LOG
    ucilog << "gui: in check_command: " << buf << (flush) << endl;
#endif
    terminate = 0;
    if (uci && strcmp(buf,"stop")==0) {
        // do not add this to the pending stack, but set the stopped
        // flag so we know we got the command.
        terminate++; stopped++;
#ifdef UCI_LOG
        ucilog << "got stop" << (flush) << endl;
#endif
        return;
    }
    else if (uci && strcmp(buf,"quit")==0) {
        add_pending(buf);
        terminate++;
        return;
    }
    else if (uci && strcmp(buf,"ponderhit")==0) {
        // We predicted the opponent's move, so we need to
        // continue doing the ponder search but adjust the time
        // limit.
        ++ponderhit;
        ponder_move_ok = true;
        ASSERT(pondering);
        // continue the search in non-ponder mode
        if (srctype != FixedDepth) {
            // Compute how much longer we must search
            ColorType side = searcher->getComputerSide();
            time_target =
                (srctype == FixedTime) ? time_limit :
                calcTimeLimit(movestogo, 
                              side == White ? winc : binc,
                              time_left, opp_time, !easy, doTrace);
            if (time_target <= 0) time_target = 30;
            if (doTrace) {
                cout << "# time_target = " << time_target << endl;
            }
            searcher->setTimeLimit(time_target,calc_extra_time(side));
        }
        searcher->setTalkLevel(Whisper);
        searcher->setBackground(false);
        // Since we have shifted to foreground mode, show the current
        // search statistics:
        post_output(ponder_stats);
    }
    else if (uci && (strncmp(buf,"position",8)==0 ||
                     strcmp(buf,"ucinewgame")==0)) {
        terminate = 1; // stop this search in order to start a new one
        add_pending(buf);
    } else if (uci) {
        // Most other commands do not terminate the search. Execute them now.
        // (technically, according the UCI spec, setoption is not allowed
        // during search: but UIs such as ChessBase assume it is).
        Board &board = pondering ? *ponder_board : *main_board;
        if (!do_command(buf,board)) {
            terminate = 1; // program is terminating
        }
    }
    else if (strcmp(buf,"quit")==0 || strcmp(buf,"end")==0 || strncmp(buf,"test",4)==0) {
        add_pending(buf);
        analyzeMode = 0;
        terminate = 1;
    }
    else if (strcmp(buf,"new")==0) {
        add_pending(buf);
        terminate = 1;
    }
    else if (strncmp(buf,"random",6)==0 || strncmp(buf, "ics ",4) == 0) {
        // ignore
    }
    else if (strcmp(buf,"computer")==0) {
        computer = 1;
    }
    else if (strcmp(buf,"?") == 0) {
        // Winboard 3.6 or higher sends this to terminate a search
        // in progress
        if (doTrace) cout << "# terminating." << endl;
        terminate = true;
    }
    else if (strncmp(buf,"ping",4) == 0) {
        // new for Winboard 4.2
        if (pondering) {
            cout << "pong " << buf+5 << (flush) << endl;
        }
        else {
            // we got a ping, probably in the process of terminating a game.
            // Execute the "ping" command only after the search completes.
            add_pending(buf);
        }
    }
    else if (strcmp(buf,"hint")==0) {
        doHint();
    }
    else if (strcmp(buf,"depth")==0) {
    }
    else if (strncmp(buf,"level",5)==0) {
        parseLevel(buf+6);
        srctype = TimeLimit;
    }
    else if (strncmp(buf,"st ",3)==0) {
        sscanf(buf+3,"%d",&time_limit);
        srctype = FixedTime;
    }
    else if (strncmp(buf,"sd ",3)==0) {
        sscanf(buf+3,"%d",&ply_limit);
        srctype = FixedDepth;
    }
    else if (strncmp(buf,"time",4)==0) {
        // my time left
        sscanf(buf+4,"%d",&time_left);
    }
    else if (strncmp(buf,"otime",4)==0) {
        // opponent's time left
        sscanf(buf+4,"%d",&opp_time);
    }
    else if (strcmp(buf,"post")==0) {
        post = 1;
    }
    else if (strcmp(buf,"nopost")==0) {
        post = 0;
    }
    else if (strcmp(buf,"savegame")==0) {
    }
    else if (strcmp(buf,"remove")==0 || strcmp(buf,"undo")==0) {
        add_pending(buf);
        terminate = true;
    }
    else if (strcmp(buf,"resign")==0 || strncmp(buf,"result",6)==0) {
        add_pending(buf);
        game_end = 1;
        if (doTrace)
            cout << "# received_result: " << buf << endl;
        terminate = 1;
    }
    else if (strcmp(buf,"draw")==0) {
        // "draw" command. Requires winboard 3.6 or higher.
        if (accept_draw(*main_board)) {
            // Notify opponent. don't assume draw is concluded yet.
            cout << "offer draw" << endl;
        }
        else if (doTrace)
            cout << "# draw declined" << endl;
    }
    else if (strcmp(buf,"edit")==0) {
        // TBD what to do in analyze mode
        add_pending(buf);
        terminate = true;
    }
    else if (strncmp(buf,"setboard",8) == 0 ||
        strcmp(buf,"analyze")==0) {
            add_pending(buf);
            terminate = true;
    }
    else if (strcmp(buf,".") == 0) {
        // for analysis mode
        cout << "stat01: " <<
            stats.elapsed_time << " " << stats.num_nodes << " " <<
            stats.depth << " " <<
            stats.mvleft << " " << stats.mvtot << endl;
    }
    else if (strcmp(buf,"go") == 0 || strcmp(buf,"exit") == 0) {
        add_pending(buf);
        terminate = true;
    }
    else if (strcmp(buf,"bk")==0) {
    }
    else if (strcmp(buf,"easy")==0) {
        easy = true;
    }
    else if (strcmp(buf,"hard")==0) {
        easy = false;
    }
    else if (strcmp(buf,"white")==0 || strcmp(buf,"black")==0 || strcmp(buf,"playother")==0) {
        add_pending(buf);
        terminate = true;
    }
    else if (strncmp(buf,"name",4)==0) {
        // We've received the name of our opponent.
        opponent_name = buf+5;
    }
    else if (strncmp(buf,"bogus",5)==0) {
    }
    else if (strncmp(buf,"force",5)==0) {
        forceMode = 1;
        terminate = true;
    }
    else if (strncmp(buf,"rating",6) == 0) {
        if (doTrace) cout << "# rating = " << buf << endl;
        sscanf(buf+7,"%d %d",&computer_rating, &opponent_rating);
        if (searcher) searcher->setRatingDiff(opponent_rating-computer_rating);
    }
    else if (strncmp(buf,"accepted",8) == 0) {
    }
    else if (strncmp(buf,"computer",8) == 0) {
    }
    else if (strncmp(buf,"option",6) == 0) {
        processWinboardOptions(buf+6);
    }
    else if (strncmp(buf,"cores",5) == 0) {
        // defer until search completes
        add_pending(buf);
    }
    else if (strncmp(buf,"memory",6) == 0) {
        // defer until search completes
        add_pending(buf);
    }
    else if (strncmp(buf,"egtpath",7) == 0) {
        // defer until search completes
        add_pending(buf);
    }
    else {
        if (strncmp(buf,"usermove",8)==0) {
            // new for Winboard 4.2
            char *p = buf;
            char *q = buf+9;
            while (*q && !isspace(*q)) {
                *p++ = *q++;
            }
            *p = '\0';
        }
        // see if it could be a move
        if (isdigit(*buf)) {
            char *p = buf;
            while (*p && !isalpha(*p)) ++p;
            if (*p) {
                char *q = buf;
                while (*p) *q++=*p++;
            }
        }
        if (doTrace)
            cout << "# move text = " << buf << endl;
        if (is_move(*main_board,buf)) {
            // make the move on the board
            Move rmove = text_to_move(*main_board,buf);
            last_move = rmove;
            if (doTrace) {
                cout << "# predicted move = ";
                MoveImage(predicted_move,cout);
                cout << " last move = ";
                MoveImage(last_move,cout);
                cout << endl;
            }
            if (forceMode || analyzeMode || !pondering) {
                add_pending(buf);
                terminate = true;
            }
            else if (!IsNull(predicted_move) &&
                MovesEqual(predicted_move,last_move)) {
                    // make the move on the board
                    if (doTrace) {
                        cout << "# ponder ok" << endl;
                    }
                    execute_move(*main_board,last_move);
                    // We predicted the opponent's move, so we need to
                    // continue doing the ponder search but adjust the time
                    // limit.
                    ponder_move_ok = true;
                    if (srctype != FixedDepth) {
                        // Compute how much longer we must search

                        ColorType side = searcher->getComputerSide();
                        time_target =
                            (srctype == FixedTime) ? time_limit :
                            (uci ? calcTimeLimit(movestogo, 
                                                 getIncrUCI(side),
                                                 time_left, opp_time,
                                                 true, doTrace)
                             : calcTimeLimit(moves, minutes, incr, time_left, opp_time, true, doTrace));
                        if (time_target <= 0) time_target = 30;
                        if (doTrace) {
                            cout << "# time_target = " << time_target << endl;
                            cout << "# xtra time = " << calc_extra_time(side) << endl;
                        }
                        searcher->setTimeLimit(time_target,calc_extra_time(side));
                    }
                    searcher->setTalkLevel(Whisper);
                    searcher->setBackground(false);
                    post_output(ponder_stats);
                    terminate = false;
            }
            else {
                if (doTrace) cout << "# ponder not ok" << endl;
                // We can't use the results of pondering because we
                // did not predict the opponent's move.  Stop the
                // search and then execute the move.
                ponder_move_ok = false;
                add_pending(buf);
                terminate = true;
            }
        }
        else if (doTrace) {
            cout << "# warning: move string not parsed" << endl;
        }
    }
}


// get next whitespace-separated token
static int nextToken(const string &source,
                     string::const_iterator &it, string &token)
{
    token.clear();
    while (it != source.end() && isspace(*it)) it++;
    while (it != source.end() && !isspace(*it)) token += *it++;
    return token.size();
}


static const char *get_move(const char *buf,const Board &board, Move &m)
{
   const char *p = buf;
      m = NullMove;
   if (*p) {
      while (isspace(*p) && *p != '\0') ++p;
         if (*p == '\0')
         return NULL;
         char tmp[10];
         int i = 0;
         char *q = tmp;
      while (!isspace(*p) && *p != '+' && *p != '\0' && i < 10) {
         *q++ = *p++;
            ++i;
      }
      *q = '\0';
         m = Notation::value(board,board.sideToMove(),tmp);
         if (*p == '+') ++p;
   }
   return p;
}


// for support of the "test" command
static Move search(Board &board, int ply_limit,
int time_limit, Statistics &stats,
Move excludes [], int num_excludes) {
   Move move = NullMove;
   solution_time = -1;

   move = searcher->findBestMove(board,
      srctype,
      time_limit, 0, ply_limit,
      0, 0, stats,
      verbose ? Debug : Silent,
      excludes, num_excludes);

   if (num_excludes)
      cout << "result(" << num_excludes+1 << "):";
   else
      cout << "result:";
   cout << '\t';
   Notation::image(board,move,cout);
   cout << "\tscore: ";
   Scoring::printScore(stats.display_value,cout);
   cout <<  '\t';
   total_time += stats.elapsed_time;
   total_nodes += stats.num_nodes;
   gameMoves->removeAll();
   return move;
}


static void do_test(string test_file)
{
   delayedInitIfNeeded();
   int tmp = options.book.book_enabled;
   options.book.book_enabled = 0;
   total_nodes = (uint64)0;
   total_correct = total_tests = 0;
   total_time = (uint64)0;
   nodes_to_find_total = (uint64)0;
   depth_to_find_total = 0;
   time_to_find_total = (uint64)0;

   solution_times.removeAll();
   Board board;
   ifstream pos_file( test_file.c_str(), ios::in);
   if (!pos_file) {
      cout << "Failed to open EPD file." << endl;
         return;
   }
   char buf[512];
   while (!pos_file.eof()) {
      pos_file.getline(buf,511);
      if (!pos_file) {
         cout << "Error reading EPD file." << endl;
            return;
      }
      // Try to parse this line as an EPD command.
      stringstream stream(buf);
      string id, comment;
      EPDRecord epd_rec;
      if (!ChessIO::readEPDRecord(stream,board,epd_rec)) break;
      if (epd_rec.hasError()) {
         cerr << "error in EPD record ";
         if (id.length()>0) cerr << id;
         cerr << ": ";
         cerr << epd_rec.getError();
         cerr << endl;
      }
      else {
         int i;
         solution_move_count = 0;
         int illegal=0;
         id = "";
         for (i = 0; i < epd_rec.getSize(); i++) {
            string key, val;
            epd_rec.getData(i,key,val);
            if (key == "bm" || key == "am") {
               Move m;
               const char *p = val.c_str();
               while (*p) {
                  p = get_move(p,board,m);
                  if (IsNull(m)) {
                     ++illegal;
                  }
                  else if (solution_move_count < 10) {
                     solution_moves[solution_move_count++] = m;
                  }
                  avoid = (key == "am");
               }
            }
            else if (key == "id") {
               id = val;
            }
            else if (key == "c0") {
               comment = val;
            }
         }
         if (illegal) {
            cerr << "illegal or invalid solution move(s) for EPD record ";
            if (id.length()>0) cerr << id;
            cerr << endl;
            continue;
         }
         else if (!solution_move_count) {
            cerr << "no solution move(s) for EPD record ";
            if (id.length()>0) cerr << id;
            cerr << endl;
            continue;
         }
         last_iteration_depth = -1;
         iterations_correct = 0;
         early_exit = 0;
         solution_time = -1;
         max_depth = 0;
         testing = 1;
         cout << id << ' ';
         if (comment.length()) cout << comment << ' ';
         if (avoid) {
            cout << "am ";
            Notation::image(board,solution_moves[0],cout);
            cout << endl;
         }
         else {
            cout << "bm";
            for (int i = 0; i < solution_move_count; i++) {
               cout << ' ';
               Notation::image(board,solution_moves[i],cout);
            }
            cout << endl;
         }
         srctype = FixedTime;
         Move excludes[Constants::MaxMoves];
         for (int index = 0; index < moves_to_search; index++) {
            searcher->clearHashTables();
            Move result = search(board,ply_limit,time_limit,stats,
            excludes,index);
            if (IsNull(result)) break;
            excludes[index] = result;
            int correct = solution_time >=0;
            solution_times.append((int)solution_time);
            total_tests++;
            if (correct) {
               cout << "\t++ solved in " << setprecision(2) << (float)solution_time/100.0 <<
                  " sec. (";
               print_nodes(solution_nodes,cout);
               cout << " nodes)" << endl;
               total_correct++;
            }
            else {
               cout << "\t** not solved in " << setprecision(2) <<
                  (float)stats.elapsed_time/100.0 << " secs. (";
               print_nodes(stats.num_nodes,cout);
               cout << " nodes)" << endl;
            }
            cout << stats.best_line_image << endl;
            if (correct && max_depth>0) {
               uint64 nodes_to_find = (uint64)0;
               int depth_to_find = 0;
               time_t time_to_find = 0;
               for (int i=max_depth-1;i>=0;i--) {
                  if (avoid) {
                     if (MovesEqual(search_progress[i].move,result)) {
                        nodes_to_find = search_progress[i+1].num_nodes;
                        time_to_find = search_progress[i+1].time;
                        depth_to_find = search_progress[i+1].depth;
                        break;
                     }
                  }
                  else {
                     if (!solution_match(search_progress[i].move)) {
                        nodes_to_find = search_progress[i+1].num_nodes;
                        time_to_find = search_progress[i+1].time;
                        depth_to_find = search_progress[i+1].depth;
                        break;
                     }
                  }
               }
               nodes_to_find_total += nodes_to_find;
               depth_to_find_total += depth_to_find;
               time_to_find_total += time_to_find;
            }
         }
      }

      char c;
      while (!pos_file.eof()) {
         c = pos_file.get();
         if (!isspace(c) && c != '\n') {
            if (!pos_file.eof())
               pos_file.putback(c);
               break;
         }
      }
   }
   pos_file.close();
   cout << endl << "solution times:" << endl;
   cout << "         ";
   int i = 0;
   for (i = 0; i < 10; i++)
      cout << i << "      ";
   cout << endl;
   double score = 0.0;
   for (i = 0; i < solution_times.length(); i++) {
      char digits[15];
      if (i == 0) {
         sprintf(digits,"% 4d |       ",i);
         cout << endl << digits;
      }
      else if ((i+1) % 10 == 0) {
         sprintf(digits,"% 4d |",(i+1)/10);
         cout << endl << digits;
      }
      if (solution_times[i] == -1) {
         cout << "  ***  ";
      }
      else {
         sprintf(digits,"%6.2f ",solution_times[i]/100.0);
         cout << digits;
         score += (float)time_limit/100.0 - solution_times[i]/100.0;
      }
   }
   cout << endl << endl << "correct : " << total_correct << '/' <<
   total_tests << endl;
   if (total_correct) {
      string avg = "";
      if (total_correct > 1) avg = "avg. ";
      cout << avg << "nodes to solution : ";
      uint64 avg_nodes = nodes_to_find_total/total_correct;
      if (avg_nodes > 1000000L) {
         cout << (float)(avg_nodes)/1000000.0 << "M" << endl;
      }
      else
         cout << (float)(avg_nodes)/1000.0 << "K" << endl;
         cout << avg << "depth to solution : " << (float)(depth_to_find_total)/total_correct << endl;
         cout << avg << "time to solution  : " << (float)(time_to_find_total)/(100.0*total_correct) << " sec." << endl;
   }
   options.book.book_enabled = tmp;
}

// Execute a command, return false if program should terminate.
static bool do_command(const char *buf, Board &board) {
#ifdef UCI_LOG
    ucilog << "gui: " << buf << (flush) << endl;
#endif  
    if (doTrace)
        cout << "# do_command: " << buf << endl;
    if (doTrace && uci) {
        theLog->write(buf); theLog->write_eol();
    }
    if (strcmp(buf,"uci")==0) {
        uci = 1; verbose = 1;                       // TBD: fixed for now
        // Learning is disabled because we don't have full game history w/ scores
        options.learning.score_learning = options.learning.position_learning = 0;
        cout << "id name " << "Arasan " << Arasan_Version;
        cout << endl;
        cout << "id author Jon Dart" << endl;
        cout << "option name Hash type spin default " <<
            options.search.hash_table_size/(1024L*1024L) << " min 4 max 1000" << endl;
        cout << "option name Ponder type check default true" << endl;
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
        cout << "option name Use tablebases type check default ";
        if (options.search.use_tablebases) cout << "true"; else cout << "false";
        cout << endl;
        cout << "option name TablebaseType type combo";
#ifdef GAVIOTA_TBS
        cout << " var " << Options::GAVIOTA_TYPE;
#endif
#ifdef NALIMOV_TBS
        cout << " var " << Options::NALIMOV_TYPE;
#endif
        cout << " default " << options.search.tablebase_type;
        cout << endl;
#ifdef GAVIOTA_TBS
        cout << "option name GaviotaTbPath type string default " << 
            options.search.gtb_path << endl;
        cout << "option name GaviotaTbCache type spin default " << 
            options.search.gtb_cache_size/(1024*1024) << 
            " min 1 max 32" << endl;
#endif
#ifdef NALIMOV_TBS
        cout << "option name NalimovPath type string default " << 
            options.search.nalimov_path << endl;
        cout << "option name NalimovCache type spin default " <<
            options.search.nalimov_cache_size/(1024*1024) << 
            " min 1 max 32" << endl;
#endif
#endif
        cout << "option name MultiPV type spin default 1 min 1 max " << MAX_PV << endl;
        cout << "option name OwnBook type check default true" << endl;
        cout << "option name Book selectivity type spin default " <<
            options.book.selectivity << " min 0 max 100" << endl;
        cout << "option name Threads type spin default " <<
            options.search.ncpus << " min 1 max " <<
            Constants::MaxCPUs << endl;
        cout << "uciok" << endl;
        return true;
    }
    else if (strcmp(buf,"quit")==0) {
        return false;
    }
    else if (uci && strncmp(buf,"setoption",9)==0) {
        string args(buf+9);
        string name, value;
        size_t nam = args.find("name");
        if (nam != string::npos) {
            // search for "value"
            size_t val = args.find(" value",nam+4);
            if (val != string::npos) {
               name = args.substr(nam+4,val-nam-4);
               // trim spaces
               name = name.erase(0 , name.find_first_not_of(' ') );
               name = name.erase( name.find_last_not_of(' ') + 1);
               value = args.substr(val+6);
               value = value.erase(0, value.find_first_not_of(' '));
               value = value.erase(value.find_last_not_of(' ') + 1);
            }
        }
        if (name == "Hash") {
            size_t old = options.search.hash_table_size;
            // size is in megabytes
            istringstream buf(value);
            int size;
            buf >> size;
            options.search.hash_table_size = (size_t)size*1024L*1024L;
            if (old != options.search.hash_table_size) {
                Hash::resizeHash(options.search.hash_table_size);
            }
        }
        else if (name == "Ponder") {
            easy = !(value == "true");
        }
#if defined(NALIMOV_TBS) || defined(GAVIOTA_TBS)
        else if (name == "Use tablebases") {
            options.search.use_tablebases = (value == "true");
        }
        else if (name == "Tablebase type") {
#ifdef GAVIOTA_TBS
            if (value == Options::GAVIOTA_TYPE) {
                options.search.tablebase_type = Options::GAVIOTA_TYPE;
            }
#endif
#ifdef NALIMOV_TBS
            if (value == Options::NALIMOV_TYPE) {
                options.search.tablebase_type = Options::NALIMOV_TYPE;
            }
#endif
        }
#endif
#ifdef NALIMOV_TBS
        else if (name == "NalimovPath") {
            options.search.nalimov_path = value;
            options.search.use_tablebases = 1;
            options.search.tablebase_type = Options::NALIMOV_TYPE;
        }
        else if (name == "NalimovCache") {
            // This is in MB
            int size;
            if (setIntOption(value,size)) {
                options.search.nalimov_cache_size = Util::Min(32,size)*1024L*1024L;
            }
        }
#endif
#ifdef GAVIOTA_TBS
        else if (name == "GaviotaTbPath") {
            options.search.gtb_path = value;
            options.search.use_tablebases = 1;
            options.search.tablebase_type = Options::GAVIOTA_TYPE;
        }
        else if (name == "GaviotaTbCache") {
            // This is in MB
            int size;
            if (setIntOption(value,size)) {
                options.search.gtb_cache_size = Util::Min(32,size)*1024L*1024L;
            }
        }
#endif
        else if (name == "OwnBook") {
            options.book.book_enabled = (value == "true");
        }
        else if (name == "Book selectivity") {
            setIntOption(value,options.book.selectivity);
        }
        else if (name == "MultiPV") {
            setIntOption(value,options.search.multipv);
            options.search.multipv = Util::Min(MAX_PV,options.search.multipv);
            // GUIs (Shredder at least) send 0 to turn multi-pv off: but
            // our option counts the # of lines to show, so we set it to
            // 1 in this case.
            if (options.search.multipv == 0) options.search.multipv = 1;
            stats.multipv_count = 0;
            // migrate current stats to 1st Multi-PV table entry:
            multi_pvs[0] = MultiPVEntry(stats);
        }
        else if (name == "Threads") {
            int threads = options.search.ncpus;
            if (setIntOption(value,threads) && threads >0 && threads <= Constants::MaxCPUs) {
                options.search.ncpus = threads;
                searcher->setThreadCount(options.search.ncpus);
            }
        }
    }
    else if (uci && strcmp(buf,"ucinewgame")==0) {
        do_command("new",board);
        return true;
    }
    else if (uci && strcmp(buf,"isready")==0) {
        delayedInitIfNeeded();
        cout << "readyok" << endl;
#ifdef UCI_LOG
        ucilog << "readyok" << endl;
#endif
    }
    else if (uci && strncmp(buf,"position",8)==0) {
        ponder_move = NullMove;
        const char *p = buf+9;
        const char *pos_start = NULL;
        if (strncmp(p,"startpos",8)==0) {
            pos_start = p+9;
            board.reset();
            gameMoves->removeAll();
        }
        else if (strncmp(p,"fen",3)==0) {
            char fen[256]; char *q = fen; p+=4;
            pos_start = p;
            while (*p && *p != '\013' && q-fen<255) {
                *q++=*p++;
            }
            *q = '\0';
            if (BoardIO::readFEN(board, fen)) {
                if (doTrace) cout << "# read fen" << endl;
#ifdef UCI_LOG
                ucilog << "read FEN" << endl;
#endif
            }
            else {
                if (doTrace) cout << "# warning: invalid fen!" << endl;
#ifdef UCI_LOG
                ucilog << "warning: invalid FEN!" << endl;
#endif
            }
        }
        const char *moves;
        if (pos_start && (moves = strstr(pos_start,"moves"))!=NULL) {
            // "moves" clause
            p = moves+6;
            while (*p) {
              while (*p && isspace(*p)) ++p;
              char movebuf[64]; char *q = movebuf;
              while (*p && !isspace(*p)) *q++=*p++;
              *q = '\0';
              if (strlen(movebuf)==0) break;
              Square start = SquareValue(movebuf);
              Square dest = SquareValue(movebuf+2);
              PieceType promotion;
              switch (*(movebuf+4)) {
                 case 'q': promotion = Queen; break;
                 case 'n': promotion = Knight; break;
                 case 'b': promotion = Bishop; break;
                 case 'r': promotion = Rook; break;
                 default: promotion = Empty; break;
              }
              Move m = CreateMove(board,start,dest,promotion);
              BoardState previous_state = board.state;
              board.doMove(m);
              gameMoves->add_move(board,previous_state,m,"");
            }
        }
    }
    else if (strncmp(buf,"test",4)==0) {
        string filename;
        ofstream *out_file = NULL;
        streambuf *sbuf = cout.rdbuf();
        char *p = strtok((char*)(buf+5)," ");
        if (p) {
            filename = p;
            p = strtok(NULL," ");
            if (p) {
                time_limit = 100*atoi(p);
                p = strtok(NULL," ");
                early_exit_plies = Constants::MaxPly;
                moves_to_search = 1;
                verbose = 0;
                while (p) {
                    if (strcmp(p,"-v")==0) {
                        p = strtok(NULL," "); 
                       ++verbose;
                        continue;
                    }
                    else if (strcmp(p,"-x")==0) {
                        p = strtok(NULL," ");
                        early_exit_plies = atoi(p);
                        p = strtok(NULL," ");
                    }
                    else if (strcmp(p,"-N")==0) {
                        p = strtok(NULL," ");
                        moves_to_search = atoi(p);
                        p = strtok(NULL," ");
                    }
                    else {
                        out_file = new ofstream( p, ios::out | ios::trunc);
                        // redirect stdout
                        cout.rdbuf(out_file->rdbuf());
                        break;
                    }
                }
                do_command("new",board);
                PostFunction old_post = searcher->registerPostFunction(post_test);
                TerminateFunction old_terminate = searcher->registerTerminateFunction(terminate);
                Options tmp = options;
                options.book.book_enabled = 0;
                options.learning.position_learning = 0;
                options.learning.score_learning = 0;
                options.learning.result_learning = 0;
                do_test(filename);
                if (out_file) {
                    out_file->close();
                    delete out_file;
                    cout.rdbuf(sbuf);               // restore console output
                }
                options = tmp;
                searcher->registerPostFunction(old_post);
                searcher->registerTerminateFunction(old_terminate);
                cout << "test complete" << endl;
            }
            else
                cout << "invalid command" << endl;
        }
        else
            cout << "invalid command" << endl;
    }
    else if (strncmp(buf,"eval",4)==0) {
        string filename;
        const char *p = buf + 4;
        while (p && isspace(*p)) p++;
        if (p) {
            filename = p;
            ifstream pos_file( p, ios::in);
            pos_file >> board;
            if (!pos_file.good()) {
                cout << "File not found, or bad format." << endl;
            }
            else {
                delayedInitIfNeeded();
#if defined(NALIMOV_TBS) || defined(GAVIOTA_TBS)
                int tbscore;
                if (options.search.use_tablebases) {
#ifdef GAVIOTA_TBS
                    if (options.search.tablebase_type == 
                        Options::GAVIOTA_TYPE && GaviotaTb::probe_tb(board,tbscore,0,1)) {
                        cout << "score = ";
                        Scoring::printScore(tbscore,cout);
                        cout << " (from " << options.search.tablebase_type << " tablebases)" << endl;
                    }
#endif
#ifdef NALIMOV_TBS
                    if (options.search.tablebase_type == 
                        Options::NALIMOV_TYPE && NalimovTb::probe_tb(board,tbscore,0)) {
                        cout << "score = ";
                        Scoring::printScore(tbscore,cout);
                        cout << " (from " << options.search.tablebase_type << " tablebases)" << endl;
                    }
#endif
                }
#endif
                int score;
                if ((score = Scoring::tryBitbase(board))!= Scoring::INVALID_SCORE) {
                    cout << "bitbase score=";
                    Scoring::printScore(score,cout);
                    cout << endl;
                }
                Scoring::init();
                SearchController c;
                if (Scoring::isDraw(board))
                    cout << "position evaluates to draw (statically)" << endl;
                Scoring *s = new Scoring();
                s->init();
                cout << board << endl;
                Scoring::printScore(s->evalu8(board),cout);
                cout << endl;
                board.flip();
                cout << board << endl;
                Scoring::printScore(s->evalu8(board),cout);
                delete s;
                cout << endl;
            }
        }
    }
    else if (uci && strcmp(buf,"stop")==0) {
        ++stopped;
        return true;
    }
    else if (uci && strncmp(buf,"go",2)==0) {
        string option, value;
        srctype = TimeLimit;
        int do_ponder = 0;
        movestogo = 0;
        bool infinite = false;
        string cmds(buf+2);
        string::const_iterator it = cmds.begin();
        while (it != cmds.end()) {
            if (!nextToken(cmds,it,option)) break;
            if (option == "wtime") {
                srctype = TimeLimit;
                if (!nextToken(cmds,it,value)) break;
                int t;
                if (setIntOption(value,t)) {
                   if (board.sideToMove() == White)
                       time_left = t/10;
                   else
                       opp_time = t/10;
                }
            }
            else if (option == "btime") {
                srctype = TimeLimit;
                if (!nextToken(cmds,it,value)) break;
                int t;
                if (setIntOption(value,t)) {
                    if (board.sideToMove() == Black)
                        time_left = t/10;
                    else
                        opp_time = t/10;
                }
            }
            else if (option == "infinite") {
                srctype = FixedTime;
                time_limit = INFINITE_TIME;
                infinite = true;
            }
            else if (option == "winc") {
                if (!nextToken(cmds,it,value)) break;
                // incr is in milliseconds
                int tmp;
                if (setIntOption(value,tmp)) {
                    winc = tmp/10;
                }
            }
            else if (option == "binc") {
                if (!nextToken(cmds,it,value)) break;
                // incr is in milliseconds
                int tmp;
                if (setIntOption(value,tmp)) {
                    binc = tmp/10;
                }
            }
            else if (option == "movestogo") {
                if (!nextToken(cmds,it,value)) break;
                setIntOption(value,movestogo);
            }
            else if (option == "depth") {
                if (!nextToken(cmds,it,value)) break;
                srctype = FixedDepth;
                setIntOption(value,ply_limit);
            }
            else if (option == "movetime") {
                if (!nextToken(cmds,it,value)) break;
                srctype = FixedTime;
                int t;
                if (setIntOption(value,t)) {
                    time_limit = t/10;
                }
            }
            else if (option == "ponder") {
                ++do_ponder;
                ponderhit = 0;
            }
        }
        stopped = 0;
        forceMode = 0;
        if (do_ponder) {
            ponder(board,NullMove,NullMove,1);
            // We should not send the move unless we have received a
            // "ponderhit" command, in which case we were pondering the
            // right move. If pondering completes early, we must wait
            // for ponderhit before sending. But also send the move if
            // we were stopped - this is the "handshake" that tells the
            // UI we received the stop.
#ifdef UCI_LOG
            ucilog << "done pondering: stopped=" << stopped << (flush) << endl;
#endif  
            if (ponderhit || stopped) {
                // ensure we send an "info" command - may not have been
                // sent if the previous move was forced or a tb hit.
                uciOut(ponder_stats);
                send_move(board,ponder_move,ponder_stats);
                ponder_move = NullMove;
                ponderhit = 0;
                stopped = 0;
            }
            else {
                // We completed pondering early - the protocol requires
                // that we delay sending the move until "ponderhit" or
                // "stop" is received
                uciWaitState = 1;
            }
        }
        else {
#ifdef UCI_LOG
            ucilog << "starting search" << (flush) << endl;
#endif  
            best_move = search(searcher,board,stats,infinite);

#ifdef UCI_LOG
            ucilog << "done searching: stopped=" << stopped << (flush) << endl;
#endif  
            if (options.search.multipv == 1) {
               uciOut(stats);
            }
            if (infinite && !stopped) {
                // We were told "go infinite" but completed searching early
                // (due to a mate or forced move or tb hit). The protocol 
                // requires that we go into a wait state before sending the
                // move. We will exit when a "stop" command is received.
                uciWaitState = 1;
            }
            else {
                send_move(board,best_move,stats);
            }
            stopped = 0;
        }
    }
    else if (uci && strcmp(buf,"ponderhit")==0) {
        ++ponderhit;
        if (!IsNull(ponder_move)) {
            uciOut(ponder_stats);
            send_move(board,ponder_move,ponder_stats);
        }
    }
    else if (strcmp(buf,"help")==0) {
        do_help();
    }
    else if (strcmp(buf,"end")==0) {
        return false;
    }
    else if (strcmp(buf,"new")==0) {
        save_game();
        board.reset();
        theLog->clear();
        if (!uci) theLog->write_header();
        computer_plays_white = false;
        // Note: "new" does not reset analyze mode
        forceMode = 0;
        game_end = 0;
        testing = 0;
        computer = 0;
        opponent_name = "";
        stats.clear();
        ponder_stats.clear();
        last_stats.clear();
        last_move = NullMove;
        last_move_image.clear();
        gameMoves->removeAll();
        predicted_move = NullMove;
        pondering = 0;
        ponder_move_ok = false;
        start_fen.clear();
        searcher->registerCommandFunction(check_command);
        searcher->registerPostFunction(post_output);
        delayedInitIfNeeded();
        searcher->clearHashTables();
        if (ics)
            cout << "kib Hello from Arasan " << Arasan_Version << endl;
    }
    else if (strncmp(buf,"random",6)==0) {
        // ignore
    }
    else if (strncmp(buf,"protover",8)==0) {
        // new in Winboard 4.2
        cout << "feature name=1 setboard=1 san=1 usermove=1 ping=1 ics=1 playother=0 sigint=0 colors=0 analyze=1 debug=1 memory=1 smp=1";
#if defined(GAVIOTA_TBS) && defined(NALIMOV_TBS)
        cout << " egt=\"gaviota,nalimov\" ";
#elif defined(GAVIOTA_TBS)
        cout << " egt=\"gaviota\" ";
#elif defined(NALIMOV_TBS)
        cout << " egt=\"nalimov\" ";
#endif
        // custom option for book selectivity
        cout << " option=\"Book selectivity -spin " <<
            options.book.selectivity << " 1 100\"";
        cout << " option=\"Can resign -check " <<
            options.search.can_resign << "\"";
        cout << " option=\"Result learning -check " <<
            options.learning.result_learning << "\"";
        cout << " option=\"Position learning -check " <<
            options.learning.position_learning << "\"";
        cout << " option=\"Score learning -check " <<
            options.learning.score_learning << "\"";
        cout << " myname=\"" << "Arasan " << Arasan_Version << "\"" << endl;
        ++xboard42;
        // set done = 0 because it may take some time to initialize tablebases.
        cout << "feature done=0" << endl;
        delayedInitIfNeeded();
        cout << "feature done=1" << endl;
#ifdef UCI_LOG
        ucilog << "feature done=1" << endl;
#endif
    }
    else if (strcmp(buf,"computer")==0) {
        computer = 1;
    }
    else if (strncmp(buf,"ping",4) == 0) {
        cout << "pong " << buf+5 << (flush) << endl;
    }
    else if (strncmp(buf,"ics",3) == 0) {
        hostname = "";
        for (const char *p=buf+4;*p;) {
            hostname += *p++;
        }
    }
    else if (strcmp(buf,"hint")==0) {
        doHint();
    }
    else if (strcmp(buf,"bk")==0) {
        Move moves[20];
        int scores[20];
        int count = 0;
        delayedInitIfNeeded(); // to allow "bk" before "new"
        if (options.book.book_enabled && openingBook) {
            count = openingBook->book_moves(*main_board,moves,scores,20);
        }
        if (count == 0) {
            cout << '\t' << "No book moves for this position." << endl
                << endl;
        }
        else {
            cout << " book moves:" << endl;
            for (int i = 0; i<count; i++) {
                cout << '\t';
                Notation::image(*main_board,moves[i],cout);
                cout << endl;
            }
            cout << endl;
        }
    }
    else if (strcmp(buf,"depth")==0) {
    }
    else if (strncmp(buf,"level",5)==0) {
        parseLevel(buf+6);
        srctype = TimeLimit;
    }
    else if (strncmp(buf,"st ",3)==0) {
        sscanf(buf+3,"%d",&time_limit);
        srctype = FixedTime;
    }
    else if (strncmp(buf,"sd ",3)==0) {
        sscanf(buf+3,"%d",&ply_limit);
        srctype = FixedDepth;
    }
    else if (strncmp(buf,"time",4)==0) {
        // my time left
        sscanf(buf+4,"%d",&time_left);
    }
    else if (strncmp(buf,"otime",4)==0) {
        sscanf(buf+4,"%d",&opp_time);
    }
    else if (strcmp(buf,"post")==0) {
        post = 1;
    }
    else if (strcmp(buf,"nopost")==0) {
        post = 0;
    }
    else if (strncmp(buf,"result",6)==0) {
        // Game has ended
        theLog->setResult(buf+7);
        save_game();
        game_end = 1;
        gameMoves->removeAll();
    }
    else if (strcmp(buf,"savegame")==0) {
    }
    else if (strcmp(buf,"remove")==0) {
        undo(board);
        undo(board);
    }
    else if (strcmp(buf,"undo")==0) {
        undo(board);
    }
    else if (strcmp(buf,"resign")==0) {
        // our opponent has resigned
        cout << "# setting log result" << endl;
        if (computer_plays_white)
            theLog->setResult("0-1");
        else
            theLog->setResult("1-0");
        cout << "# set log result" << endl;
    }
    else if (strcmp(buf,"draw")==0) {
        // "draw" command. Requires winboard 3.6 or higher.
        if (accept_draw(board)) {
            // Notify opponent. don't assume draw is concluded yet.
            cout << "offer draw" << endl;
        }
        else if (doTrace)
            cout << "# draw declined" << endl;
    }
    else if (strncmp(buf,"setboard",8) == 0) {
        start_fen = buf+8;
        start_fen = start_fen.erase(0 , start_fen.find_first_not_of(' ') );
        stringstream s(start_fen,ios::in);
        ChessIO::load_fen(s,board);
    }
    else if (strcmp(buf,"edit")==0) {
        edit_board(cin,board);
    }
    else if (strcmp(buf,"analyze")==0) {
        analyzeMode = 1;
        analyze(board);
    }
    else if (strcmp(buf,"exit")==0) {
        if (analyzeMode) analyzeMode = 0;
    }
    else if (strcmp(buf,".") == 0) {
        // for analysis mode: not supported presently
    }
    else if (strcmp(buf,"go")==0) {
        forceMode = 0;
        // set the side flag here - do not rely on the deprecated
        // "white" and "black" commands.
        computer_plays_white = board.sideToMove() == White;
        Move reply = search(searcher,board,stats,false);
        if (!forceMode) send_move(board,reply,stats);
    }
    else if (strcmp(buf,"bk")==0) {
    }
    else if (strcmp(buf,"easy")==0) {
        easy = true;
    }
    else if (strcmp(buf,"hard")==0) {
        easy = false;
    }
    else if (strcmp(buf,"white")==0) {
        computer_plays_white = true;
        //forceMode = 0;
    }
    else if (strcmp(buf,"black")==0) {
        computer_plays_white = false;
        //forceMode = 0;
    }
    /**   else if (strcmp(buf,"playother")==0) {
    computer_plays_white = !computer_plays_white;
    forceMode = 0;
    int tmp = opp_time;
    opp_time = time_left;
    time_left = tmp;
    }**/
    else if (strncmp(buf,"name",4)==0) {
        // We've received the name of our opponent.
        opponent_name = buf+5;
    }
    else if (strncmp(buf,"bogus",5)==0) {
    }
    else if (strncmp(buf,"force",5)==0) {
        forceMode = 1;
    }
    else if (strncmp(buf,"rating",6) == 0) {
        if (doTrace) cout << "# rating = " << buf << endl;
        sscanf(buf+7,"%d %d",&computer_rating, &opponent_rating);
        if (searcher) searcher->setRatingDiff(opponent_rating-computer_rating);
    }
    else if (strncmp(buf,"accepted",8) == 0) {
    }
    else if (strncmp(buf,"computer",8) == 0) {
    }
    else if (strncmp(buf,"option",6) == 0) {
        // Winboard option command
        processWinboardOptions(buf+6);
    }
    else if (strncmp(buf,"cores",5) == 0) {
        // Setting -c on the Arasan command line takes precedence over
        // what the GUI sets
        if (!cpusSet) {
           // set number of threads
           string cmd(buf);
           stringstream ss(cmd.substr(6));
           if((ss >> options.search.ncpus).fail()) {
               cerr << "invalid value following 'cores'" << endl; 
           } else {
               options.search.ncpus = Util::Min(options.search.ncpus,Constants::MaxCPUs);
               searcher->setThreadCount(options.search.ncpus);
           }
        }
    }
    else if (strncmp(buf,"memory",6) == 0) {
        // Setting -H on the Arasan command line takes precedence over
        // what the GUI sets
        if (!memorySet) {
           // set memory size in MB
           string cmd(buf);
           stringstream ss(cmd.substr(7));
           int mbs;
           if((ss >> mbs).fail()) {
               cerr << "invalid value following 'memory'" << endl; 
           } else {
               options.search.hash_table_size = (size_t)(mbs*1024L*1024L);
               Hash::freeHash();
               Hash::initHash(options.search.hash_table_size);
           }
        }
    } 
#ifdef NALIMOV_TBS
    else if  (strncmp(buf,"egtpath nalimov",15) == 0) {
        // tablebase path. Note: setting this after "new" is not 
        // supported. We require an engine restart for it to be
        // effective.
        options.search.use_tablebases = 1;
        options.search.nalimov_path = string(buf+16);
#ifdef _WIN32
        // path may be in Unix format. Convert.
        string::iterator it = options.search.nalimov_path.begin();
        while (it != options.search.nalimov_path.end()) {
            if (*it == '/') {
                *it = '\\';
            }
            it++;
        }
#endif        
        if (doTrace) {
            cout << "# setting Nalimov tb path to " << options.search.nalimov_path << endl;
        }
    }
#endif
#ifdef GAVIOTA_TBS
    else if  (strncmp(buf,"egtpath gaviota",15) == 0) {
        // tablebase path. Note: setting this after "new" is not 
        // supported. We require an engine restart for it to be
        // effective.
        options.search.use_tablebases = 1;
        options.search.gtb_path = string(buf+16);
#ifdef _WIN32
        // path may be in Unix format. Convert.
        string::iterator it = options.search.gtb_path.begin();
        while (it != options.search.gtb_path.end()) {
            if (*it == '/') {
                *it = '\\';
            }
            it++;
        }
#endif        
        if (doTrace) {
            cout << "# setting Gaviota tb path to " << options.search.gtb_path << endl;
        }
    }
#endif
    else {
        // see if it could be a move
        char movetext[80];
        if (strncmp(buf,"usermove",8) == 0) {
            // new for Winboard 4.2
            strcpy(movetext,buf+9);
        }
        else
            strcpy(movetext,buf);
        if (isdigit(*movetext)) {
            const char *p = movetext;
            while (*p && !isalpha(*p)) ++p;
            if (*p) {
                char *q = movetext;
                while (*p && !isspace(*p)) *q++=*p++;
                *q = '\0';
            }
        }
        if (doTrace)
            cout << "# move text = " << movetext << endl;
        Move rmove;
process_move:
        if (is_move(board,movetext)) {
            if (game_end) {
                if (forceMode)
                    game_end = 0;
                else {
                    if (doTrace) cout << "# ignoring move " << movetext << " received after game end" << endl;
                    return true;
                }
            }
            if (doTrace) {
                cout << "# got move: " << movetext << endl;
            }
            // assume move
            //cout << 1+gameMoves->num_moves()/2 << ". " << movetext << endl;

            rmove = text_to_move(board,movetext);
            if (!IsNull(rmove)) {
                // make the move on the board
                execute_move(board,rmove);
                if (analyzeMode) {
                    // re-enter analysis loop
                    analyze(board);
                }
                Move reply;
                if (!forceMode && !analyzeMode) {
                    // determine what to do with the pondering result, if
                    // there is one.
                    if (MovesEqual(predicted_move,rmove) && !IsNull(ponder_move)) {
                        // We completed pondering already and we got a reply to
                        // this move (e.g. might be a forced reply).
                        if (doTrace) cout << "# pondering complete already" << endl;
                        if (doTrace) {
                            cout << "# sending ponder move ";
                            MoveImage(ponder_move,cout);
                            cout << endl << (flush);
                        }
                        reply = ponder_move;
                        stats = ponder_stats;
                        post_output(stats);
                        game_end |= stats.end_of_game;
                        if (doTrace) cout << "# game_end = " << game_end << endl;
                        predicted_move = ponder_move = NullMove;
                    }
                    else {
                        predicted_move = ponder_move = NullMove;
                        reply = search(searcher,board,stats,false);
                        // Note: we may know the game has ended here before
                        // we get confirmation from Winboard. So be sure
                        // we set the global game_end flag here so that we won't
                        // start pondering after the game is over.
                        game_end |= stats.end_of_game;
                        if (doTrace) {
                            cout << "# state = " << stats.state << endl;
                            cout << "# game_end = " << game_end  << endl;
                        }
                    }
                    // Check for game end conditions like resign, draw acceptance, etc:
                    if (check_pending(board)==1)
                        game_end = true;
                    else if (!forceMode) {
                        // call send_move even if game_end = true because we
                        // handle resignation, etc. there.
                        send_move(board,reply,stats);
                    }
                }
ponder:
                if (!forceMode && !analyzeMode && !game_end && !easy && time_target >= 100 /* 1 second */) {
                    // check pending commands again before pondering in case we have a
                    // resign or draw, or a move has come in (no good pondering if the
                    // opponent has replied already).
                    ponder_move_ok = false;
                    int result;
                    if ((result = check_pending(board)) == 0) {
                        ponder(board,reply,stats.best_line[1],uci);
                        if (check_pending(board)==1) {
                            return 1;                       // game end signal seen
                        }
                    }
                    else if (result == 1)                 // game end
                        return 1;
                    // We are done pondering. If we got a ponder hit (opponent made
                    // our predicted move) we are ready to move now
                    if (ponder_move_ok && !IsNull(ponder_move) && !game_end
                        && !forceMode && !analyzeMode) {
                            // we got a reply from pondering
                            if (doTrace)
                                cout << "# sending ponder move" << endl;
                            stats = ponder_stats;
                            send_move(board,ponder_move,stats);
                            post_output(stats);
                            reply = ponder_move;
                            predicted_move = ponder_move = NullMove;
                            goto ponder;
                    }
                    else if (num_pending && check_pending(board)==2) {
                        if (doTrace) cout << "# ponder failed, move in stack" << endl;
                        ponder_move = NullMove;
                        const char *cmd = pending[0].c_str();
                        if (strncmp(cmd,"usermove",8)==0) {
                            strcpy(movetext,cmd+9);
                        }
                        else {
                            strcpy(movetext,cmd);
                        }
                        // remove from pending list
                        for (int j=1;j<num_pending;j++) pending[j-1] = pending[j];
                        --num_pending;
                        goto process_move;
                    }
                }
            }
        }
    }
    return true;
}


int CDECL main(int argc, char **argv) {
    signal(SIGINT,SIG_IGN);

    // Show a message on the console
    cout << "Arasan " Arasan_Version << ' ' << Arasan_Copyright << endl;
    // Must use unbuffered console
    setbuf(stdin,NULL);
    setbuf(stdout, NULL);
    std::cout.rdbuf()->pubsetbuf(NULL, 0);
    std::cin.rdbuf()->pubsetbuf(NULL, 0);

    Bitboard::init();
    initOptions(argv[0]);
    Attacks::init();
    Scoring::init();
    if (!initGlobals(argv[0], true)) {
        cleanupGlobals();
        exit(-1);
    }
    atexit(cleanupGlobals);

#ifdef UCI_LOG
    ucilog.open(derivePath("ucilog").c_str(),ios::out|ios::app);
    ucilog << "starting up" << endl;
#endif

#ifdef _WIN32
    // setup polling thread for input from engine
    polling_terminated = 0;
    DWORD id;
    pollingThreadHandle = CreateThread(NULL,0,
        inputPoll,NULL,
        0,
        &id);
#endif

    ecoCoder = new ECO();
    Board board;
    main_board = &board;
    int arg = 1;

    if (argc > 1) {
        while (arg < argc && *(argv[arg]) == '-') {
            char c = *(argv[arg]+1);
            switch (c) {
            case 'c':
                ++arg;
                options.search.ncpus = Util::Min(Constants::MaxCPUs,atol(argv[arg]));
                ++cpusSet;
                if (options.search.ncpus<=0) {
                    cerr << "-c parameter must be >=1" << endl;
                    exit(-1);
                }
                break;
            case 'i':
                if (strcmp(argv[arg]+1,"ics")==0)
                    ics = true;
                else {
                    cerr << "Warning: unknown option: " << argv[arg]+1 <<
                        endl;
                }
                break;
            case 'H':
                ++arg;
                Options::setMemoryOption(options.search.hash_table_size,
                    string(argv[arg]));
                ++memorySet;
                break;
            case 't':
                //++arg;
                doTrace = true;
                break;
            default:
                cerr << "Warning: unknown option: " << argv[arg]+1 <<
                    endl;
                break;
            }
            ++arg;
        }
    }
    if (arg < argc) {
        cout << "loading " << argv[arg] << endl;
        ifstream pos_file( argv[arg], ios::in);
        if (pos_file.good()) {
            pos_file >> board;
        }
        else {
            cout << "file not found: " << argv[arg] << endl;
            return -1;
        }
    }

    last_move_image.clear();

    if (options.store_games) {
        if (options.game_pathname == "") {
            game_pathname = derivePath("games.pgn");
        }
        else {
            game_pathname = options.game_pathname;
        }
        game_file = new ofstream(game_pathname.c_str(),ios::out | ios::app);
        if (!game_file->good()) {
            cerr << "Warning: cannot open game file. Games will not be saved." << endl;
            delete game_file;
            game_file = NULL;
        }
    }
    Hash::initHash((size_t)(options.search.hash_table_size));

    searcher = new SearchController();

    if (testing) {
        do_test(test_file);
    }
    else {
        searcher->registerCommandFunction(check_command);
        searcher->registerPostFunction(post_output);

#ifdef _WIN32
        while(!polling_terminated) {
            // do a timed wait so we can terminate thread w/no input 
            WaitForSingleObject(inputEvent,2000);
            Lock(input_lock);
            // got some input, maybe
            char *lasteol = NULL;
            // find last newline. Can't use strrchr because might not be
            // null terminated
            if (cmd_buf_count) {
                char *p = cmd_buf+cmd_buf_count-1;
                while (p != cmd_buf && *p != '\n') --p;
                if (*p == '\n') {
                    lasteol = p;
                }
            }
            char buf2[CMD_BUF_SIZE];
            *buf2 = '\0';
            if (lasteol != NULL) {
                char *unread = lasteol + 1;
                // copy command line so we can unlock buffer
                int line_length = (int)(unread-cmd_buf);
                ASSERT(line_length>=0 &&
                    line_length<=CMD_BUF_SIZE);
                if (line_length>0) {
                    strncpy(buf2,cmd_buf,line_length);
                }
                *(buf2+line_length) = '\0';
                // remove characters we copied from cmd_buf
                ASSERT(cmd_buf_count-line_length>=0 &&
                    cmd_buf_count-line_length<=CMD_BUF_SIZE);
                if(cmd_buf_count-line_length>0) {
                    strncpy(cmd_buf,unread,cmd_buf_count-line_length);
                }
                cmd_buf_count -= line_length;
            }
            Unlock(input_lock);
            char *p = buf2;
            while (*p) {
                while (*p == '\n' || *p == '\015') p++;
                string cmd_line = "";
                while (*p && *p != '\n' && *p != '\015') {
                    cmd_line += *p++;
                }
                if (*p) {
                    if (doTrace) cout << "# got cmd (main): " << cmd_line.c_str() << " len=" << cmd_line.length() << endl;
#ifdef UCI_LOG
                    ucilog << "calling do_command(main): " << cmd_line << (flush) << endl;
#endif
                    if (uciWaitState) {
#ifdef UCI_LOG
                        ucilog << "got command in wait state: " << cmd_line << (flush) << endl;
#endif
                        // we expect a "stop" or "ponderhit"
                        if (cmd_line == "ponderhit") {
#ifdef UCI_LOG
                            ucilog << "received ponderhit" << endl;
#endif
                            send_move(board,ponder_move,ponder_stats);
                            uciWaitState = 0;
                        } else if (cmd_line == "stop") {
#ifdef UCI_LOG
                            ucilog << "received stop" << endl;
#endif
                            send_move(board,best_move,stats);
                            uciWaitState = 0;
                        } else if (cmd_line == "quit") {
#ifdef UCI_LOG
                            ucilog << "received quit" << endl;
#endif
                            polling_terminated++;
                            break;
                        }
                    }
                    else {
#ifdef UCI_LOG
                        ucilog << "calling do_command(main):" << cmd_line << (flush) << endl;
#endif
                        if (!do_command(cmd_line.c_str(),board)) {
                            polling_terminated++;
                            break;
                        }
                        if (!do_pending(board)) {
                            polling_terminated++;
                            break;
                        }
                    }
                    ++p;
                }
            }
        }
#else // not WIN32
        // Read stdin until we get a quit command
        int nextc;
        while ((nextc = getc(stdin)) != EOF) {
            if (nextc == '\n') {
                cmd_buf[cmd_buf_count] = '\0';
                cmd_buf_count = '\0';
                if (doTrace) cout << "# got cmd (main): " << cmd_buf << endl;
                if (!do_command(cmd_buf,board))
                    break;
                if (!do_pending(board))
                    break;
            }
            else {
                if (cmd_buf_count >= CMD_BUF_SIZE-1) {
                    cout << "# warning: command buffer full" << endl;
                    cmd_buf_count = 0;
                }
                cmd_buf[cmd_buf_count++] = nextc;
            }
        }
#endif
        // handle termination.

        save_game();
    }
    if (doTrace) cout << "# terminating" << endl;
    delete searcher;
    delete ecoCoder;
#ifdef _WIN32
    polling_terminated++;
    WaitForSingleObject(pollingThreadHandle,1000);
#endif
    delete game_file;
#ifdef UCI_LOG
    ucilog.close();
#endif
    return 0;
}
